from functools import wraps

from ..fixture import rules_combine, build_chain


def make_rule(rules=None):
    return rules_combine(*build_chain(rules or []))


def make_both_rules(rules=None):
    _rule = make_rule(rules)

    @wraps(_rule)
    def _plain_rule(*args, **kwargs):
        _value, _ = _rule(*args, **kwargs)
        return _value

    return _rule, _plain_rule



import copy
from contextlib import contextmanager
from functools import lru_cache
from itertools import chain
from typing import Tuple, Optional

from .imports import ImportPool, ImportStatement, FromImport, DirectImport, try_import_info


@lru_cache()
def _get_unprocessable():
    from .common import unprocessable
    return unprocessable


def unprocessable():
    return _get_unprocessable()()


class AddonProxy:
    """
    Overview:
        Proxy of addon expression.
    """

    def __init__(self, base, rule):
        """
        Overview:
            Constructor of ``AddonProxy``

        Arguments:
            - base: Base unit of the proxy.
            - rule: Translation rule.
        """
        self.__base = base
        self.__rule = rule

    def __getattr__(self, item: str):
        """
        Overview:
            Generate attribute-getting expression, \
            like ``base.xxx``.

        Arguments:
            - item (:obj:`str`): Attribute name.

        Returns:
            - proxy: New generated proxy.
        """
        return AddonProxy(f'{self.__base}.{item}', self.__rule)

    def __getitem__(self, item):
        """
        Overview:
            Generate item-getting expression, \
            like ``base[233]``, ``base[::-1]`` or ``base[1, 2, 3]``.

        Arguments:
            - item (:obj:`str`): Item value, can be simple value, slice or tuple.

        Returns:
            - proxy: New generated proxy.
        """
        if isinstance(item, slice):
            _items = [item.start, item.stop, *((item.step,) if item.step is not None else ())]
            _item_str = ':'.join(map(lambda x: '' if x is None else self.__rule(x), _items))
        elif isinstance(item, tuple) and len(item) > 0:
            _item_str = ', '.join(map(self.__rule, item))
            _item_str = _item_str + (', ' if len(item) == 1 else '')
        else:
            _item_str = self.__rule(item)
        return AddonProxy(f'{self.__base}[{_item_str}]', self.__rule)

    def __call__(self, *args, **kwargs):
        """
        Overview:
            Generate item call expression, \
            like ``base(1, 2, x=1, y=2)``.

        Arguments:
            - args: Positional arguments.
            - kwargs: Key-word arguments.

        Returns:
            - proxy: New generated proxy.
        """
        _items = chain(
            map(lambda x: self.__rule(x), args),
            map(lambda p: f"{p[0]}={self.__rule(p[1])}", kwargs.items())
        )
        return AddonProxy(
            f'{self.__base}({", ".join(_items)})',
            self.__rule,
        )

    def __str__(self) -> str:
        """
        Overview:
            String format of expression.

        Returns:
            - return (:obj:`str`): String formatted expression.
        """
        return self.__base


class Addons:
    """
    Overview:
        Addon object used in rules.
    """

    def __init__(self, rule=None):
        """
        Overview:
            Constructor of ``Addons``.

        Arguments:
            - rule: Translation rule, default is ``None`` which means no translation rule, \
                auto translation will also be disabled when ``None`` is used.
        """
        self.__imports = ImportPool()
        self.__rule = rule

    def import_(self, _import: str) -> DirectImport:
        return self.__imports.import_(_import)

    def from_(self, _from: str) -> FromImport:
        return self.__imports.from_(_from)

    def obj(self, obj, alias: Optional[str] = None) -> AddonProxy:
        """
        Overview:
            Generate an object proxy.
            Object proxy means a object which need to be automatically imported, \
            such as ``json``, an import statement ``import json`` will be generated.
            If the object cannot be imported, it will raise exception.

        Arguments:
            - obj: Target object.
            - alias (:obj:`Optional[str]`): Alias name of the imported object, \
                default is ``None`` which means do not alias.

        Returns:
            - proxy (:obj:`AddonProxy`): Expression proxy object.
        """
        with self.transaction():
            segments = try_import_info(obj, alias=alias)
            _import = self
            for sign, content in zip(segments[::2], segments[1::2]):
                if sign == 'from':
                    _import = _import.from_(content)
                elif sign == 'import':
                    _import = _import.import_(content)
                else:
                    _import = _import.as_(content)

            if alias:
                _import = _import.as_(alias)
            return AddonProxy(str(_import.target), self.rule)

    def val(self, obj) -> AddonProxy:
        """
        Overview:
            Generate a value proxy.
            Differs from ``obj`` method, this method will not do automatic import, \
            it will just generate the string expression with the rule translator.

        Arguments:
            - obj: Value object.

        Returns:
            - proxy (:obj:`AddonProxy`): Expression proxy object.
        """
        with self.transaction():
            return AddonProxy(self.rule(obj), self.rule)

    # noinspection PyMethodMayBeStatic
    def raw(self, val, *, func=None) -> AddonProxy:
        """
        Overview:
            Generate raw proxy.

        Arguments:
            - val: Raw object.
            - func: Translation function, default is ``None`` which means just \
                turn it to string by builtin function ``str``.

        Returns:
            - proxy (:obj:`AddonProxy`): Raw proxy object.
        """
        return AddonProxy((func or str)(val), self.rule)

    def rule(self, v) -> str:
        """
        Overview:
            Express the value with the rule translator.

        Arguments:
            - v: Value object.

        Returns:
            - str (:obj:`str`): String formatted expression.
        """
        with self.transaction():
            return str(self.__rule(v, self))

    @property
    def import_items(self) -> Tuple[ImportStatement, ...]:
        """
        Overview:
            Import statements.

        Returns:
            - statements (:obj:`Tuple[ImportStatement, ...]`): Tuple of import statements.
        """
        return self.__imports.imports

    # noinspection PyMethodMayBeStatic
    def unprocessable(self):
        """
        Overview:
            Raise an exception ``UnprocessableError``.
            Announce that the current rule cannot express this value.
        """
        unprocessable()

    @contextmanager
    def transaction(self):
        """
        Overview:
            Open up an atomic transaction context.
            If exception is thrown in this context, all the update in this \
            addon object will be undone.

        Example:
            >>> with addon.transaction():
            >>>     addon.rule(1)
            >>>     addon.val([1, 2])
            >>>     addon.obj(json)
            >>>     raise RuntimeError  # the changes above will be undone
        """
        _state = copy.deepcopy(self.__getstate__())
        try:
            yield self
        except Exception as err:
            self.__setstate__(_state)
            raise err

    def __getstate__(self):
        """
        Overview:
            Get state value of this object.

        Returns:
            - state: State value.
        """
        return self.__imports, self.__rule

    def __setstate__(self, state):
        """
        Overview:
            Set state value of this object.

        Arguments:
            - state: State value.
        """
        self.__imports, self.__rule = state



from typing import Tuple, List

import click
from hbutils.reflection import quick_import_object

from .base import CONTEXT_SETTINGS, _is_rule_block
from .utils import validator, err_validator, multiple_validator, rules_struct
from ...translate import transobj


@err_validator((ImportError,))
@multiple_validator
@validator
def validate_rules(value: str):
    rule, _, _ = quick_import_object(value, _is_rule_block)
    return rule


@err_validator((ImportError,))
@validator
def validate_input(value: str):
    return quick_import_object(value)


def _trans_cli(cli: click.Group):
    @cli.command('trans', help='Translate existing python object to visible code.',
                 context_settings=CONTEXT_SETTINGS)
    @click.option('-r', '--rules', 'rules', type=click.UNPROCESSED, callback=validate_rules, multiple=True,
                  default=[], help='External rules definition.')
    @click.option('-I', '--information', is_flag=True, default=False,
                  help='Show detailed information of imported object.')
    @click.argument('input_', type=click.UNPROCESSED, callback=validate_input)
    def _trans(input_: Tuple[object, str, str], rules: List[object], information: bool):
        _obj, _module, _name = input_
        _type = type(_obj)
        _result = transobj(_obj, trans=rules)

        if information:
            click.secho(click.style('Object Information:', fg='blue'))
            if _module:
                click.secho(f'Module : {_module}')
            click.secho(f'Name : {_name}')
            click.secho(f'Type : {_type.__qualname__}')
            if rules:
                click.secho('External Rules :')
                click.secho(transobj(rules, trans=rules_struct).code)
            click.secho()

            import_list = sorted(set(_result.imports), key=lambda x: x.key)
            if import_list:
                click.secho(click.style('Import Statements:', fg='cyan'))
                for imp in import_list:
                    click.secho(str(imp))
                click.secho()

            click.secho(click.style('Expression:', fg='green'))
            click.secho(_result)
            click.secho()
        else:
            click.secho(_result)

    return cli



from .base import _base_potc_cli
from .export import _export_cli
from .trans import _trans_cli
from .utils import _cli_builder

potc_cli = _cli_builder(
    _base_potc_cli,
    _trans_cli,
    _export_cli,
)



from .builtin import builtin_all, builtin_reflect, builtin_collection, builtin_basic, builtin_ellipsis, \
    builtin_slice, builtin_range, builtin_bytes, builtin_float, builtin_module, \
    builtin_tuple, builtin_func, builtin_type, builtin_dict, builtin_int, builtin_list, builtin_none, builtin_set, \
    builtin_str, builtin_items, builtin_complex, builtin_object
from .typing import typing_wrapper, typing_all, typing_typevar, typing_callable, typing_items

native_all = [
    typing_all,
    builtin_all,
]



import re
from collections import namedtuple
from functools import wraps
from typing import Callable, Type, Union, Tuple, Optional

from hbutils.reflection import dynamic_call


class UnprocessableError(Exception):
    """
    Overview:
        Signal exception which means the rules before cannot process this data.
    """
    pass


def unprocessable():
    """
    Overview:
        Raise an ``UnprocessableError``.
    """
    raise UnprocessableError()


RuleInfo = namedtuple('RuleInfo', ['name', 'origin'])

_RULE_TAG = '__rule__'
_NAME_PATTERN = re.compile('^[a-zA-Z0-9_]+$')


def rule(alias: Optional[str] = None, type_: Union[Type, Tuple[Type, ...], None] = None):
    """
    Overview:
        Make a common function to a rule function.

    Arguments:
        - alias (:obj:`Optional[str]`): Alias name of this rule, default is ``None`` which means \
            just use the name of the wrapped function.
        - type\\_ (:obj:`Union[Type, Tuple[Type, ...], None]`): Type of the data, can be a type \
            or a tuple of types, default is ``None`` which means no type limit.

    Returns:
        - decorator: A function decorator of the raw rule function.

    Examples:
        There are some example from builtin part.

        - builtin_int (A very easy case)

        >>> @rule(type_=int)
        >>> def builtin_int(v: int):
        >>>     return repr(v)

        - builtin_float (A little complex case)

        >>> @rule(type_=float)
        >>> def builtin_float(v: float, addon: Addons):
        >>>     if math.isinf(v):
        >>>         return ('+' if v > 0 else '-') + str(addon.obj(math).inf)
        >>>     elif math.isnan(v):
        >>>         return addon.obj(math).nan
        >>>     else:
        >>>         if math.isclose(v, math.e):
        >>>             return addon.obj(math).e
        >>>         elif math.isclose(v, math.pi):
        >>>             return addon.obj(math).pi
        >>>         elif math.isclose(v, math.tau):
        >>>             return addon.obj(math).tau
        >>>         else:
        >>>             return repr(v)

    """

    def _decorator(func: Callable):
        if is_rule(func):
            _name = alias or rule_name(func)
            return rule(_name, type_)(rule_origin(func))
        else:
            _name = alias or func.__name__
            _actual_func = dynamic_call(func)
            if not _NAME_PATTERN.fullmatch(_name):
                raise NameError(f'Invalid rule name, '
                                f'{repr(_NAME_PATTERN.pattern)} expected but {repr(_name)} found.')

            @wraps(func)
            def _new_func(v, addon):
                if type_ is not None and not isinstance(v, type_):
                    unprocessable()
                with addon.transaction():
                    return _actual_func(v, addon)

            setattr(_new_func, _RULE_TAG, RuleInfo(_name, func))
            return _new_func

    return _decorator


def is_rule(func: Callable) -> bool:
    """
    Overview:
        Check if a function is a potc rule.

    Arguments:
        - func (:obj:`Callable`): Function object.

    Returns:
        - is_rule (:obj:`bool`): Is a rule or not.

    Examples:
        >>> is_rule(builtin_int)            # True
        >>> is_rule(lambda x, addon: None)  # False
    """
    return isinstance(getattr(func, _RULE_TAG, None), RuleInfo)


def rule_name(rule_: Callable) -> str:
    """
    Overview:
        Name of the rule.

    Arguments:
        - rule\\_ (:obj:`Callable`): Rule object.

    Returns:
        - name (:obj:`str`): Name of the rule.

    Examples:
        >>> rule_name(builtin_int)  # 'builtin_int'
    """
    return getattr(rule_, _RULE_TAG).name


def rule_origin(rule_: Callable) -> Callable:
    """
    Overview:
        Original function of the rule.

    Arguments:
        - rule\\_ (:obj:`Callable`): Rule object.

    Returns:
        - func (:obj:`str`): Original function.
    """
    return getattr(rule_, _RULE_TAG).origin



import zlib

import dill


def dump_obj(obj: object) -> bytes:
    """
    Overview:
        Dump any object to binary format.
        In order to reduce the size of bytes as much as possible, \
        zlib compression is applied in this function.

    Arguments:
        - obj (:obj:`object`): Any object.

    Returns:
        - data (:obj:`bytes`): Dumped bytes data.
    """
    return zlib.compress(dill.dumps(obj))


def load_obj(data: bytes) -> object:
    """
    Overview:
        Load any object from binary format.

    Arguments:
        - data (:obj:`bytes`): Dumped bytes data.

    Returns:
        - obj (:obj:`object`): Original object.
    """
    return dill.loads(zlib.decompress(data))



import click
from click.core import Context, Option

from ...config.meta import __TITLE__, __VERSION__, __AUTHOR__, __AUTHOR_EMAIL__
from ...fixture import is_rule

_raw_authors = [item.strip() for item in __AUTHOR__.split(',') if item.strip()]
_raw_emails = [item.strip() for item in __AUTHOR_EMAIL__.split(',')]
if len(_raw_emails) < len(_raw_authors):
    _raw_emails += [None] * (len(_raw_authors) - len(_raw_emails))
elif len(_raw_emails) > len(_raw_authors):
    _raw_emails[len(_raw_authors) - 1] = tuple(_raw_emails[len(_raw_authors) - 1:])
    del _raw_emails[len(_raw_authors):]

_author_tuples = [
    (author, tuple([item for item in (email if isinstance(email, tuple) else ((email,) if email else ())) if item]))
    for author, email in zip(_raw_authors, _raw_emails)
]
_authors = [
    author if not emails else '{author} ({emails})'.format(author=author, emails=', '.join(emails))
    for author, emails in _author_tuples
]


# noinspection PyUnusedLocal
def print_version(ctx: Context, param: Option, value: bool) -> None:
    """
    Print version information of cli
    :param ctx: click context
    :param param: current parameter's metadata
    :param value: value of current parameter
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo('{title}, version {version}.'.format(title=__TITLE__.capitalize(), version=__VERSION__))
    if _authors:
        click.echo('Developed by {authors}.'.format(authors=', '.join(_authors)))
    ctx.exit()


def _is_rule_block(item):
    return isinstance(item, (list, tuple)) or is_rule(item)


CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Show package's version information.")
def _base_potc_cli():
    pass



from ..fixture import Addons, rule
from ..supports import raw_object
from ..supports.bin import dump_obj


@rule()
def default_object(v, addon: Addons):
    return addon.obj(raw_object)(addon.raw(dump_obj(v), func=repr))


#: Overview:
#:      Default object type.
#:      Used as the default implement, when all the rules are disabled, \
#:      this rule and system rules will still work to make sure the basic property.
#:
#: Items:
#:      - Default expression of :class:`object`
default_all = [
    (
        default_object,
    ),
]



import math
import types
from functools import partial
from typing import NamedTuple, NoReturn, List, Set, Dict, Tuple, Callable, Any, TypeVar

from dill import source

from .trans import transobj_assert, transvars_assert


class _LocalType:
    pass


class _MySet(set):
    pass


class _MyList(list):
    pass


class _MyTuple(tuple):
    pass


class _MyNamedTuple(NamedTuple):
    first: object
    second: object


class _MyDict(dict):
    pass


class _MyPair:
    def __init__(self, x, y):
        self.__x = x
        self.__y = y

    def __getstate__(self):
        return self.__x, self.__y

    def __setstate__(self, state):
        self.__x, self.__y = state

    def __eq__(self, other):
        return (other is self) or (type(other) == type(self) and self.__getstate__() == other.__getstate__())


class _WrapperClass:
    class _MyInnerPair(_MyPair):
        pass

    class _MyInnerPair(_MyPair):
        pass


def provement(trans=None):
    """
    Overview:
        Create a test class containing the tests of the basic properties.

    Arguments:
        - trans: Translator object to be used.

    Examples:

        >>> import pytest
        >>> from potc_dict.plugin import __rules__  # load rules from plugin potc_dict
        >>>
        >>> from potc.testing import provement
        >>> from potc.translate import BlankTranslator
        >>>
        >>>
        >>> @pytest.mark.unittest
        >>> class TestPlugin(provement(BlankTranslator(__rules__))):
        ...     def test_pretty_dict(self):
        ...         with self.transobj_assert({'a': 1}) as (obj, name):
        ...             assert obj == {'a': 1}
        ...             assert name == 'pretty_dict'
        ...         with self.transobj_assert({}) as (obj, name):
        ...             assert obj == {}
        ...             assert name == 'pretty_dict'
        ...         with self.transobj_assert({1: 2}) as (obj, name):
        ...             assert obj == {1: 2}
        ...             assert name == 'builtin_dict'
        ...
        ...         with self.transvars_assert({'a': {'b': 1}}) as (vars_, code):
        ...             assert vars_ == {'a': {'b': 1}}
        ...             assert 'dict(b=1)' in code

    """

    class _ProvementMeta(type):
        def __init__(cls, name, basis, attrs):
            type.__init__(cls, name, basis, attrs)
            cls.transobj_assert = partial(transobj_assert, trans=trans)
            cls.transvars_assert = partial(transvars_assert, trans=trans)

    # noinspection PyMethodMayBeStatic
    class _Provement(metaclass=_ProvementMeta):
        def test_str(self):
            with self.transobj_assert('') as (obj, name):
                assert obj == ''
            with self.transobj_assert('str') as (obj, name):
                assert obj == 'str'
            with self.transobj_assert('str' * 1000) as (obj, name):
                assert obj == 'str' * 1000

        def test_int(self):
            with self.transobj_assert(0) as (obj, name):
                assert obj == 0
            with self.transobj_assert(123) as (obj, name):
                assert obj == 123
            with self.transobj_assert(-3249) as (obj, name):
                assert obj == -3249

        def test_float(self):
            with self.transobj_assert(123.0) as (obj, name):
                assert obj == 123.0

            with self.transobj_assert(float('nan')) as (obj, name):
                assert math.isnan(obj)
            with self.transobj_assert(float('+inf')) as (obj, name):
                assert obj == math.inf
            with self.transobj_assert(float('-inf')) as (obj, name):
                assert obj == -math.inf

            with self.transobj_assert(math.e) as (obj, name):
                assert obj == math.e
            with self.transobj_assert(math.pi) as (obj, name):
                assert obj == math.pi
            with self.transobj_assert(math.tau) as (obj, name):
                assert obj == math.tau

        def test_none(self):
            with self.transobj_assert(None) as (obj, name):
                assert obj is None

        def test_bytes(self):
            with self.transobj_assert(b'') as (obj, name):
                assert obj == b''
                assert isinstance(obj, bytes)
            with self.transobj_assert(b'lsdjkflpisdjgp89erdjpo9\x00\x11') as (obj, name):
                assert obj == b'lsdjkflpisdjgp89erdjpo9\x00\x11'
                assert isinstance(obj, bytes)

            with self.transobj_assert(bytearray()) as (obj, name):
                assert obj == bytearray()
                assert isinstance(obj, bytearray)
            with self.transobj_assert(bytearray(b'lsdjkflpisdjgp89erdjpo9\x00\x11')) as (obj, name):
                assert obj == bytearray(b'lsdjkflpisdjgp89erdjpo9\x00\x11')
                assert isinstance(obj, bytearray)

        def test_type(self):
            with self.transobj_assert(int) as (obj, name):
                assert obj == int

            with self.transobj_assert(_LocalType) as (obj, name):
                assert obj == _LocalType

            with self.transobj_assert(_WrapperClass._MyInnerPair) as (obj, name):
                assert obj == _WrapperClass._MyInnerPair

            with self.transobj_assert(types.ModuleType) as (obj, name):
                assert obj is types.ModuleType

        def test_func(self):
            with self.transobj_assert(lambda x: x ** x) as (obj, name):
                assert obj(2) == 4
                assert obj(3) == 27

        def test_list(self):
            with self.transobj_assert([1, 2, 3]) as (obj, name):
                assert obj == [1, 2, 3]

            with self.transobj_assert([1, 'klsdfj', -1.15, lambda x: x ** x]) as (obj, name):
                assert obj[:3] == [1, 'klsdfj', -1.15]
                assert obj[3](2) == 4
                assert obj[3](3) == 27

            with self.transobj_assert([]) as (obj, name):
                assert obj == []

            with self.transobj_assert([1]) as (obj, name):
                assert obj == [1]

            with self.transobj_assert(_MyList([1, 2, 3])) as (obj, name):
                assert isinstance(obj, _MyList)
                assert obj == _MyList([1, 2, 3])

        def test_tuple(self):
            with self.transobj_assert(()) as (obj, name):
                assert obj == ()

            with self.transobj_assert((1,)) as (obj, name):
                assert obj == (1,)

            with self.transobj_assert((1, 2, 3)) as (obj, name):
                assert obj == (1, 2, 3)

            with self.transobj_assert((1, 'klsdfj', -1.15, lambda x: x ** x)) as (obj, name):
                assert obj[:3] == (1, 'klsdfj', -1.15)
                assert obj[3](2) == 4
                assert obj[3](3) == 27

            with transobj_assert(_MyNamedTuple(1, 2)) as (obj, name):
                assert isinstance(obj, _MyNamedTuple)
                assert obj == _MyNamedTuple(1, 2)

            with self.transobj_assert(_MyTuple((1, 2, 3))) as (obj, name):
                assert isinstance(obj, _MyTuple)
                assert obj == _MyTuple((1, 2, 3))

        def test_set(self):
            with self.transobj_assert(set()) as (obj, name):
                assert isinstance(obj, set)
                assert obj == set()

            with self.transobj_assert({1, 2, 3}) as (obj, name):
                assert isinstance(obj, set)
                assert obj == {1, 2, 3}

            with self.transobj_assert(_MySet({1, 2, 3})) as (obj, name):
                assert isinstance(obj, _MySet)
                assert obj == _MySet({1, 2, 3})

            with self.transobj_assert(frozenset({1, 2, 3})) as (obj, name):
                assert isinstance(obj, frozenset)
                assert obj == frozenset({1, 2, 3})

        def test_dict(self):
            with self.transobj_assert({}) as (obj, name):
                assert obj == {}

            with self.transobj_assert({'a': 1}) as (obj, name):
                assert obj == {'a': 1}

            with self.transobj_assert({'a': 1, 3: 'klsdfgj'}) as (obj, name):
                assert obj == {'a': 1, 3: 'klsdfgj'}

            with self.transobj_assert(_MyDict()) as (obj, name):
                assert isinstance(obj, _MyDict)
                assert obj == _MyDict()

            with self.transobj_assert(_MyDict(a=1)) as (obj, name):
                assert isinstance(obj, _MyDict)
                assert obj == _MyDict(a=1)

        def test_object(self):
            with self.transobj_assert(_MyPair(1, 2)) as (obj, name):
                assert obj == _MyPair(1, 2)

            with self.transobj_assert(_WrapperClass._MyInnerPair(1, 2)) as (obj, name):
                assert obj == _WrapperClass._MyInnerPair(1, 2)

        def test_module(self):
            with self.transobj_assert(source) as (obj, name):
                assert obj is source

        def test_ellipsis(self):
            with self.transobj_assert(...) as (obj, name):
                assert obj is Ellipsis

        def test_range(self):
            with self.transobj_assert(range(10)) as (obj, name):
                assert obj == range(10)

            with self.transobj_assert(range(-1, 6)) as (obj, name):
                assert obj == range(-1, 6)

            with self.transobj_assert(range(-1, 6, 2)) as (obj, name):
                assert obj == range(-1, 6, 2)

        def test_slice(self):
            with self.transobj_assert(slice(10)) as (obj, name):
                assert obj == slice(10)

            with self.transobj_assert(slice(-5, 10)) as (obj, name):
                assert obj == slice(-5, 10)

            with self.transobj_assert(slice(-5, 10, -4)) as (obj, name):
                assert obj == slice(-5, 10, -4)

        def test_builtin_items(self):
            with self.transobj_assert(min) as (obj, name):
                assert obj is min
                assert obj(1, 2, -1) == -1

            with self.transobj_assert(NotImplemented) as (obj, name):
                assert obj is NotImplemented

        def test_builtin_complex(self):
            with self.transobj_assert(2 + 3j) as (obj, name):
                assert obj == 2 + 3j

            with self.transobj_assert(-4j) as (obj, name):
                assert obj == 0 - 4j

            with self.transobj_assert(2 - 0j) as (obj, name):
                assert obj == 2 - 0j

        def test_mixed_case(self):
            with self.transobj_assert({
                'a': 1,
                1: {'ab', 1, 2, (), (1, 'dkls')},
                'feat': [None, type, int, 233, 'sdfkl', (5, 6, -3 + 1j), {}, set(), []],
            }) as (obj, name):
                assert obj == {
                    'a': 1,
                    1: {'ab', 1, 2, (), (1, 'dkls')},
                    'feat': [None, type, int, 233, 'sdfkl', (5, 6, -3 + 1j), {}, set(), []],
                }

        def test_typing(self):
            with transobj_assert(NoReturn) as (obj, name):
                assert obj is NoReturn

            with transobj_assert(List) as (obj, name):
                assert obj == List
            with transobj_assert(Dict) as (obj, name):
                assert obj == Dict
            with transobj_assert(Set) as (obj, name):
                assert obj == Set
            with transobj_assert(Tuple) as (obj, name):
                assert obj == Tuple
            with transobj_assert(Callable) as (obj, name):
                assert obj == Callable

            with transobj_assert(List[int]) as (obj, name):
                assert obj == List[int]
            with transobj_assert(Dict[int, List[str]]) as (obj, name):
                assert obj == Dict[int, List[str]]
            with transobj_assert(Set[str]) as (obj, name):
                assert obj == Set[str]
            with transobj_assert(Tuple[int, str, List[str]]) as (obj, name):
                assert obj == Tuple[int, str, List[str]]
            with transobj_assert(Callable[..., int]) as (obj, name):
                assert obj == Callable[..., int]
            with transobj_assert(Callable[[], int]) as (obj, name):
                assert obj == Callable[[], int]
            with transobj_assert(Callable[[int, Any], str]) as (obj, name):
                assert obj == Callable[[int, Any], str]
            with transobj_assert(Callable[[int], Any]) as (obj, name):
                assert obj == Callable[[int], Any]

            try:
                _ = list[int]
            except TypeError:
                pass
            else:
                with transobj_assert(list[int]) as (obj, name):
                    assert obj == list[int]
                with transobj_assert(dict[int, list[int]]) as (obj, name):
                    assert obj == dict[int, list[int]]

            K = TypeVar('K', bound=int, contravariant=True)
            V = TypeVar('V', int, str, covariant=True)

            with transobj_assert(K) as (obj, name):
                assert isinstance(obj, TypeVar)
                assert obj.__name__ == 'K'
                assert obj.__constraints__ == ()
                assert obj.__bound__ == int
                assert not obj.__covariant__
                assert obj.__contravariant__

            with transobj_assert(V) as (obj, name):
                assert isinstance(obj, TypeVar)
                assert obj.__name__ == 'V'
                assert obj.__constraints__ == (int, str)
                assert obj.__bound__ is None
                assert obj.__covariant__
                assert not obj.__contravariant__

            with transobj_assert(Dict[K, V]) as (obj, name):
                assert obj.__origin__ in {Dict, dict}
                _k, _v = obj.__args__
                assert isinstance(_k, TypeVar)
                assert _k.__name__ == 'K'
                assert _k.__constraints__ == ()
                assert _k.__bound__ == int
                assert not _k.__covariant__
                assert _k.__contravariant__

                assert isinstance(_v, TypeVar)
                assert _v.__name__ == 'V'
                assert _v.__constraints__ == (int, str)
                assert _v.__bound__ is None
                assert _v.__covariant__
                assert not _v.__contravariant__

    return _Provement



import warnings
from itertools import chain
from string import Template
from typing import List, Tuple, Union, Iterable

import click
from hbutils.reflection import quick_import_object, iter_import_objects

from .base import CONTEXT_SETTINGS, _is_rule_block
from .utils import validator, err_validator, multiple_validator
from ...translate import transvars


@err_validator((ImportError,))
@multiple_validator
@validator
def validate_rules(value: str):
    rule, _, _ = quick_import_object(value, _is_rule_block)
    return rule


@err_validator((ImportError, ValueError,))
@multiple_validator
@validator
def validate_value(value: str):
    try:
        name, target = value.split('=', maxsplit=1)
    except ValueError:
        name, target = '$name', value

    result = []
    np = Template(name)
    for _object, _, _full_name in iter_import_objects(target):
        _name = _full_name[::-1].split('.', maxsplit=1)[0][::-1]
        result.append((np.safe_substitute(dict(name=_name)), _object))

    return result


def _is_reformatter(item):
    return isinstance(item, (dict, str)) or callable(item)


@validator
def validate_reformat(value: str):
    try:
        _object, _, _ = quick_import_object(value, _is_reformatter)
    except ImportError:
        return value
    else:
        return _object


def _export_cli(cli: click.Group):
    @cli.command('export', help='Export existing python items to runnable source code.',
                 context_settings=CONTEXT_SETTINGS)
    @click.option('-r', '--rules', 'rules', type=click.UNPROCESSED, callback=validate_rules, multiple=True,
                  default=[], help='External rules definition.')
    @click.option('-v', '--value', 'items', type=click.UNPROCESSED, callback=validate_value, multiple=True,
                  default=None, help='Compress algorithm, can be a single lib name or a tuple, default use \'zlib\'.')
    @click.option('-f', '--reformat', 'reformat', type=click.UNPROCESSED, callback=validate_reformat, show_default=True,
                  default='google', help='Code reformatter afterwards.')
    @click.option('-F', '--no-reformat', 'no_reformat', is_flag=True, default=False,
                  help='No reformat to the exported code, -f option will be ignored.')
    def _export(items: List[Iterable[Tuple[str, object]]], rules: List[object],
                reformat: Union[str, object], no_reformat: bool):
        items = list(chain(*items))
        dvars = {}
        for name, obj in items:
            if name in dvars.keys():
                warnings.warn(UserWarning(f'Duplicate name {repr(name)} with value {repr(obj)} '
                                          f'found in -v option, and it will be ignored.'))
            else:
                dvars[name] = obj

        reformat = None if no_reformat else reformat
        _result = transvars(dvars, trans=rules, reformat=reformat)
        click.secho(_result)

    return cli



from typing import Type, TypeVar

from .bin import load_obj


def raw_object(data: bytes) -> object:
    """
    Overview:
        Raw object function support.

    Arguments:
        - data (:obj:`bytes`): Binary data.

    Returns:
        - obj (:obj:`object`): Loaded object,
    """
    return load_obj(data)


def raw_type(name: str, data: bytes) -> type:
    """
    Overview:
        Raw type function support.

    Arguments:
        - name (:obj:`str`): Name of the type.
        - data (:obj:`bytes`): Binary data.

    Returns:
        - type\_ (:obj:`object`): Loaded type,
    """
    return load_obj(data)


_T = TypeVar('_T')


# noinspection PyUnusedLocal
def typed_object(type_: Type[_T], data: bytes) -> _T:
    """
    Overview:
        Typed object function support.

    Arguments:
        - type\_ (:obj:`Type[_T]`): Type of the object.
        - data (:obj:`bytes`): Binary data.

    Returns:
        - obj (:obj:`_T`): Loaded typed object,
    """
    return load_obj(data)


def function(name: str, scheme, data: bytes):
    """
    Overview:
        Function function support.

    Arguments:
        - name (:obj:`str`): Name of the function.
        - scheme (:obj:`Type[Callable]`): Scheme of the function.
        - data (:obj:`bytes`): Binary data.

    Returns:
        - obj (:obj:`Callable`): Loaded function.
    """
    _func: scheme = load_obj(data)
    return _func
